"""Demonstrates how trading account can be set up via CMS API.
[demo] indicates that this step does not make much sense in real use case.
"""

import logging
import client.proto.CMS.common_1_pb2 as common
import client.proto.CMS.traderouting_1_pb2 as traderouting
from client.account_service import AccountService
from client.cmsapi_client import CmsApiClient
from client.profile_service import ProfileService
from client.login_service import LoginService
from config import Config

from .common import *

# Random number that is used as value when setting limits.
LIMIT_VALUE = 42

async def run(config: Config):
    print("account_setup started...")

    client = CmsApiClient()
    await client.connect(config.cms_api_url)

    logging.info("LOGON")
    await client.send_logon(
        config.username, config.password, config.client_app_id, config.private_label, config.client_version)

    # Initialize service classes.
    account_service = AccountService(client)
    profile_service = ProfileService(client)
    login_service = LoginService(client)

    logging.info("CREATE CUSTOMER PROFILE")

    # Customer profile is required to create account.
    profile = common.Profile()
    profile.legal_type = common.INDIVIDUAL
    fill_contact_information(profile.contact_information, config)
    profile.linked_brokerage_id = config.brokerage_id

    profile_id = await profile_service.create_profile(profile)

    # [demo] Get information about create profile.
    # Here and next - result is logged and not needed for anything else.
    profile = await profile_service.profile_request(profile_id)

    logging.info("CREATE ACCOUNT")

    # One of "required-create" field is `account_type_id`.
    # List of available values for this field can be obtained using next request.
    traderouting_request = traderouting.TradeRoutingRequest()
    lookup_property_list_request = traderouting_request.lookup_property_list_request
    lookup_property_list_request.property_type.append(traderouting.ACCOUNT_TYPE)
    result = await client.send_traderouting_request(traderouting_request)
    # Get ID of "NA" type (used when exchanges do not care about this value).
    account_type_id = \
        [p.value for p in result.lookup_property_list_result.lookup_property if p.name.text == "NA"][0]

    account = traderouting.Account()
    account.name = generate_name(config.brokerage_id, 64)
    account.brokerage_account_number = generate_name(config.brokerage_id, 12)
    # `class` is reserved keyword in Python.
    setattr(account, "class", traderouting.Account.Class.REGULAR)
    account.account_type_id = account_type_id
    account.profile_id = profile_id
    account.profile_sales_series_id = config.profile_sales_series_id

    account_id = await account_service.create_account(account)

    # [demo] Get general information of created account.
    account = await account_service.account_info_request(account_id)

    logging.info("UPDATE ACCOUNT SETTINGS")

    # This step is needed only when it is required to set some non-default value for one of the settings.
    # Here we enable "Instruct account" for demonstration.
    account_settings = traderouting.AccountSettings()
    account_settings.account_id = int(account_id)
    account_settings.statement_option = traderouting.AccountSettings.BROKERAGE
    account_settings.is_instruct = True

    await account_service.update_account_settings(account_settings)

    # [demo] Get updated account settings.
    account_settings = await account_service.account_settings_request(account_id)

    logging.info("UPDATE ACCOUNT RISK PARAMETERS")

    # Here we enforce couple parameters of different type and set value for them.
    # NOTE: Updating these parameters might not be possible due to brokerage restrictions.
    risk_parameters = traderouting.RiskParameters()

    # The value represented as significand * (10 ^ exponent)
    risk_parameters.enforce_trade_size_limit = True
    risk_parameters.trade_size_limit.significand = LIMIT_VALUE
    risk_parameters.trade_size_limit.exponent = 0

    # Margin Multiplier has LimitValueDouble type.
    risk_parameters.purchasing_power_parameters.margin_multiplier.mode = traderouting.LIMITED
    risk_parameters.purchasing_power_parameters.margin_multiplier.value = LIMIT_VALUE

    await account_service.update_account_risk_parameters(account_id, risk_parameters)

    # [demo] Get updated parameters.
    risk_parameters = await account_service.account_risk_parameters_request(account_id)

    logging.info("UPDATE ACCOUNT MARKET LIMITS")

    # Allow to trade all available commodities.
    # For more complicated scenarios see market_limits.py use case.
    market_limits = traderouting.MarketLimits()
    market_limits.all_market_limits.allowed_to_trade.allowed_to_trade = True

    await account_service.update_account_market_limits(account_id, market_limits)

    # [demo] Get updated market limits.
    market_limits = await account_service.account_market_limits_request(account_id)

    logging.info("CREATE TRADEROUTING LOGIN")

    # In this context "user" is historical name for "login".
    login = common.User()
    login.user_name = generate_name(config.brokerage_id, 32)
    login.subscriber_type = common.User.PRO
    login.domain = common.CQG_TRADE_ROUTING
    login.profile_id = profile_id

    login_id = await login_service.create_login(login)

    # [demo] Get general information about created login (user).
    login = await login_service.login_info_request(login_id)

    logging.info("AUTHORIZE LOGIN TO THE ACCOUNT")

    account_user_link = traderouting.AccountUserLink()
    account_user_link.account_id = int(account_id)
    account_user_link.user_id = login_id

    await account_service.authorize_login(account_user_link)

    # [demo] Get list of login authorazed to the account.
    account_login_authorization_list = await account_service.authorization_list_request(account_id)

    await client.disconnect()
    print("account_setup finished.")
