"""Demonstrates how request status can be tracked via CMS API.
"""

import logging
import time
import client.proto.CMS.cmsapi_1_pb2 as cmsapi
import client.proto.CMS.common_1_pb2 as common
import client.proto.CMS.traderouting_1_pb2 as traderouting
from client.profile_service import ProfileService
from client.cmsapi_client import CmsApiClient
from config import Config


from .common import *

async def run(config: Config):
    print("flow_tracking started...")

    client = CmsApiClient()
    await client.connect(config.cms_api_url)

    logging.info("LOGON")

    # Subscribe on request status change during logon.
    await client.send_logon(
        config.username, config.password, config.client_app_id, config.private_label, config.client_version,
        subscribe_on_request_status_change = True)

    profile_service = ProfileService(client)
    logging.info("CREATE CUSTOMER PROFILE")

    # Send customer profile request to demonstrate status change.
    profile = common.Profile()
    profile.legal_type = common.INDIVIDUAL
    fill_contact_information(profile.contact_information, config)
    profile.linked_brokerage_id = config.brokerage_id

    operation_req = cmsapi.OperationRequest()

    operation_req.create_profile.profile.CopyFrom(profile)
    operation_req.id = client._get_request_id()

    client_message: Message = cmsapi.ClientMessage()
    client_message.operation_request.append(operation_req)

    await client.send_client_message(client_message)
    operation_status = None
    
    # Check operation status, final states are SUCCESS, FAILURE, CANCELED.
    #    request flow examples: 
    #       ACCEPTED => IN_PROCESSING => SUCCESS
    #       ACCEPTED => QUEUED=> IN_PROCESSING => SUCCESS
    #    user can cancel deferred request, if needed: ACCEPTED => QUEUED => CANCELED

    while operation_status not in [common.SUCCESS, common.FAILURE, common.CANCELED]:
        result = await client.wait_for_result(operation_req)
        operation_status = result.operation_status
        print(f"Operation status: {operation_status} {_operationStatusToStr(operation_status)}")

        if operation_status == common.QUEUED:
            # Cancel request by tracking_number, if it was deferred in Waiting Queue
            cancel_message: Message = cmsapi.ClientMessage()
            cancel_req: Message = cancel_message.cancel_deferred_request
            cancel_req.tracking_number = result.tracking_number
            await client.send_client_message(cancel_message)

    await client.disconnect()
    print("flow_tracking finished.")

def _operationStatusToStr(operation_status):
    if operation_status == common.SUCCESS:
        return "SUCCESS"
    if operation_status == common.FAILURE:
        return "FAILURE"
    if operation_status == common.ACCEPTED:
        return "ACCEPTED"
    if operation_status == common.IN_PROCESSING:
        return "IN_PROCESSING"
    if operation_status == common.QUEUED:
        return "QUEUED"
    if operation_status == common.CANCELED:
        return "CANCELED"
    raise ValueError('Undefined operation_status: {}'.format(operation_status))