"""Use case that demonstrates setup of an account market limits.
For details on account creation, see account_setup case.

Steps:
1. Create customer profile to create an account

2. Create an account.

3. Enable route for the account - this defines what commodities are available for configuration for this account.

4. Get commodities that are available for the account - normally this step is required only for development,
to obtain IDs of entities and to make sure that account routes are configured correctly.

5. Update market limits for the account - the main idea of this step is to demonstrate the structure of market limits and
handling of different types of limits. There are many more limits that can be configured with this operation (see the protocol).
Please note, that data and values of the limits are for demonstration only and they might not make any sense in a real scenario.

6. Get market limits for the account - this step is for demonstration of the result market limits structure (see logs).
It is also recommended to check Account > Market Limits widget in CAST2 for visual representation of the structure.

NOTE: it is worth to mention, that in some cases, market limits message can be built using CAST2 UI,
by doing required changes in UI and then using account_market_limits_request to get the structure,
which with small adjustments can be used for another account.
"""

import logging
import client.proto.CMS.common_1_pb2 as common
import client.proto.CMS.traderouting_1_pb2 as traderouting
from client.account_service import AccountService
from client.cmsapi_client import CmsApiClient
from client.profile_service import ProfileService
from config import Config

from .common import *

async def run(config: Config):
    print("market_limits started...")

    client = CmsApiClient()
    await client.connect(config.cms_api_url)

    logging.info("LOGON")
    await client.send_logon(
        config.username, config.password, config.client_app_id, config.private_label, config.client_version)

    # Initialize service classes.
    account_service = AccountService(client)
    profile_service = ProfileService(client)

    logging.info("CREATE CUSTOMER PROFILE")

    profile = common.Profile()
    profile.legal_type = common.INDIVIDUAL
    fill_contact_information(profile.contact_information, config)
    profile.linked_brokerage_id = config.brokerage_id

    profile_id = await profile_service.create_profile(profile)

    logging.info("CREATE ACCOUNT")

    account = traderouting.Account()
    account.name = generate_name(config.brokerage_id, 64)
    account.brokerage_account_number = generate_name(config.brokerage_id, 12)
    # `class` is reserved keyword in Python.
    setattr(account, "class", traderouting.Account.Class.REGULAR)
    account.account_type_id = "0"  # N/A. See account_setup for details.
    account.profile_id = profile_id
    account.profile_sales_series_id = config.profile_sales_series_id

    account_id = await account_service.create_account(account)

    logging.info("ENABLE ROUTE FOR THE ACCOUNT")
    """At this level we enable the first available route which in its turn
    defines list of commodities that are available on this account.
    Routes that are available to an account are based on the linked brokerage and configured by CQG.
    """
    available_routes = await account_service.account_available_route_list_request(account_id)
    if len(available_routes) == 0:
        raise Exception("No routes configured for the brokerage. Contact CQG to resolve.")

    route_to_enable = traderouting.AccountRouteRecord()
    route_to_enable.route_code = int(available_routes[0].route_information.route_code)
    route_to_enable.priority = 1

    update_routes = traderouting.UpdateAccountRouteList()
    update_routes.account_id = account_id
    update_routes.routes_to_set.append(route_to_enable)
    await account_service.update_account_route_list(update_routes)

    logging.info("GET COMMODITIES THAT ARE AVAILABLE FOR THE ACCOUNT")
    """At this step we are obtaining commodities that are available on the account.
    They are grouped by exchange and have list of supported instruments (the same way as in CAST UI).
    This message also contains dictionary of instrument types that can be used to understand this list.

    This message can be used to obtain CQG specific identifiers that are required when updating market limits.
    """
    result = await account_service.account_available_exchange_groups_request(account_id)
    if len(result.exchange_group) == 0:
        raise Exception("Account has no available commodities.")

    # First exchange and its first commodity are selected here arbitrary only for demonstration purpose.
    available_exchange: traderouting.ExchangeGroup = result.exchange_group[0]
    available_commodity: traderouting.FungibleCommodity = available_exchange.fungible_commodity[0]
    # available_commodity.instrument_type_id list can be mapped to these instrument types.
    instrument_types = result.instrument_type

    logging.info("UPDATE MARKET LIMITS FOR THE ACCOUNT")
    """For this step we are using list of available commodities (see previous step).
    """
    market_limits = traderouting.MarketLimits()

    # This allows to set account-wide limits.
    # Limits on this level are applied to all commodities that have "default" (aka "inherited") mode (..LimitValue.Mode).
    # NOTE: Details of setting specific limits are extracted to "_build.." methods to make general logic clearer.
    _buildMarketLimitsRecord(market_limits.all_market_limits)

    # This allows to set limits for the whole exchange and for commodities that are listed under the exchange.
    # NOTE: Commodity can be listed only under one exchange.
    exchange_limits: traderouting.ExchangeMarketLimits = market_limits.exchange_market_limits.add()
    exchange_limits.exchange_group_id = available_exchange.id

    # Limits that are set here are applied to all commodities that are listed under the exchange
    # and have "default" mode. They override account-wide limits.
    _buildMarketLimitsRecord(exchange_limits.default_market_limits)

    # Limits that are set here are applied to the specific commodity (all instruments of this commodity)
    # and override exchange and account level limits.
    commodity_limits: traderouting.CommodityMarketLimits = exchange_limits.commodity_market_limits.add()
    commodity_limits.fungible_commodity_id = available_commodity.id
    _buildCommodityMarketLimits(commodity_limits, available_commodity)

    # Limits that are set here are applied to the specific instrument of the commodity
    # and override commodity, exchange and account level limits.
    instrument_limits: traderouting.InstrumentMarketLimits = commodity_limits.instrument_market_limits.add()
    # Select first available instrument arbitrary for demonstration.
    # This ID corresponds to InstrumentType.id.
    instrument_limits.instrument_type_id = available_commodity.instrument_type_id[0]
    _buildInstrumentMarketLimits(instrument_limits)

    await account_service.update_account_market_limits(account_id, market_limits)

    logging.info("GET MARKET LIMITS FOR THE ACCOUNT")
    """This request returns only non-default values for the account.
    For example, if MarketLimits.all_market_limits.allowed_to_trade.allowed_to_trade = True (account level)
    this means that all commodities are available to trade by default,
    so only commodities that have explicit "allowed_to_trade=True/False" will be returned.
    Default value, in this case, can be set with corresponding cleared_fields.
    Same is true for other types of limits.
    """

    market_limits = await account_service.account_market_limits_request(account_id)

    await client.disconnect()
    print("market_limits finished.")


def _buildMarketLimitsRecord(market_limits: traderouting.MarketLimitsRecord):
    # Prohibit trading for all underlying commodities by default.
    market_limits.allowed_to_trade.allowed_to_trade = False

    market_limits.margin_multiplier.mode = traderouting.LIMITED
    market_limits.margin_multiplier.value = 3

    # Unlimited for all by default.
    market_limits.position_and_trade_limits.trade_price_limit_ticks.mode = traderouting.UNLIMITED


def _buildCommodityMarketLimits(
    market_limits: traderouting.CommodityMarketLimits, available_commodity: traderouting.FungibleCommodity):

    # Enable this specific commodity for trading.
    entity_allowed_to_trade = market_limits.allowed_to_trade_commodities.add()
    # Normally there is only one tradable commodity.
    # NOTE: this ID is not the same thing as available_commodity.id
    entity_allowed_to_trade.id = available_commodity.tradable_commodity[0].id
    # To set it back to default do: entity_allowed_to_trade.cleared_fields = [2]
    entity_allowed_to_trade.allowed_to_trade = True

    market_limits.margin_multiplier.mode = traderouting.LIMITED
    market_limits.margin_multiplier.value = 4

    limits = market_limits.position_and_trade_limits

    # Still unlimited because of higher level setting.
    limits.trade_price_limit_ticks.mode = traderouting.DEFAULT

    # LimitValueDecimal example
    limits.commodity_position_limit.mode = traderouting.LIMITED
    # Result value will be 4*10^1=40.
    limits.commodity_position_limit.value.significand = 4
    limits.commodity_position_limit.value.exponent = 1 # 0 by default


def _buildInstrumentMarketLimits(market_limits: traderouting.InstrumentMarketLimits):
    limits = market_limits.position_and_trade_limits

    # Limited only for this particular instrument.
    limits.trade_price_limit_ticks.mode = traderouting.LIMITED
    limits.trade_price_limit_ticks.value = 5
